package db

import (
	"encoding/binary"
	"encoding/json"
	"time"

	"git.tcp.direct/Mirrors/bitcask-mirror"
)

// AuthLog is for implementing an authentication log.
type AuthLog interface {
	newAttempt(id int, ip, useragent string, source AuthSource, good bool) error
	getLog(id int) (map[time.Time]AuthAttempt, error)
}

type AuthSource uint32

const (
	Web    AuthSource = iota // 0
	Client                   // 1
)

// AuthAttempt represents a login attempt, used for our AuthLog.
type AuthAttempt struct {
	Time      time.Time
	IP        string
	UserAgent string
	Source    AuthSource
	Good      bool
}

// secLog is an implementation of AuthLog
type secLog struct {
	DB *bitcask.Bitcask
}

// GetUserAuths opens a bitcask database at path or starts a new one, then returns a secLog and an error.
func GetUserAuths(path string) (*secLog, error) {
	casket, err := bitcask.Open(path)
	if err != nil {
		return nil, err
	}
	return &secLog{DB: casket}, nil
}

func idToUbytes(id int) []byte {
	uid := uint32(id)
	buf := make([]byte, 4)
	binary.LittleEndian.PutUint32(buf, uid)
	return buf
}

func (s *secLog) getLog(id int) (map[time.Time]AuthAttempt, error) {
	uidb := idToUbytes(id)
	var logMap = make(map[time.Time]AuthAttempt)
	if !s.DB.Has(uidb) {
		logMap = make(map[time.Time]AuthAttempt)
		return logMap, nil
	}
	oldlogbytes, err := s.DB.Get(uidb)
	if err != nil {
		return nil, err
	}
	if err := json.Unmarshal(oldlogbytes, &logMap); err != nil {
		return nil, err
	}
	return logMap, nil
}

func (s *secLog) newAttempt(id int, ip, useragent string, source AuthSource, good bool) error {
	uidb := idToUbytes(id)

	current, err := s.getLog(id)
	if err != nil {
		return err
	}

	current[time.Now()] = AuthAttempt{
		Time:      time.Now(),
		IP:        ip,
		Source:    source,
		UserAgent: useragent,
		Good:      good,
	}

	newbytes, err := json.Marshal(current)
	if err != nil {
		return err
	}

	if err := s.DB.Put(uidb, newbytes); err != nil {
		return err
	}

	return nil
}
